import asyncio
import inspect
import json
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Generic,
    Optional,
    TypeVar,
    Union,
    overload,
)

from pydantic import BaseModel, Field, ValidationError
from typing_extensions import ParamSpec, Self

from cyberchipped.components.prompt import PromptFunction
from cyberchipped.serializers import create_tool_from_type
from cyberchipped.utilities.asyncio import (
    ExposeSyncMethodsMixin,
    expose_sync_method,
    run_async,
)
from cyberchipped.utilities.jinja import (
    BaseEnvironment,
)
from cyberchipped.utilities.logging import get_logger

if TYPE_CHECKING:
    from openai.types.chat import ChatCompletion

T = TypeVar("T")

P = ParamSpec("P")


class AIFunction(BaseModel, Generic[P, T], ExposeSyncMethodsMixin):
    fn: Optional[Callable[P, T]] = None
    environment: Optional[BaseEnvironment] = None
    prompt: Optional[str] = Field(default=inspect.cleandoc("""
        Your job is to generate likely outputs for a Python function with the
        following signature and docstring:

        {{_source_code}}

        The user will provide function inputs (if any) and you must respond with
        the most likely result.

        user: The function was called with the following inputs:
        {%for (arg, value) in _arguments.items()%}
        - {{ arg }}: {{ value }}
        {% endfor %}

        What is its output?
    """))
    name: str = "FormatResponse"
    description: str = "Formats the response."
    field_name: str = "data"
    field_description: str = "The data to format."
    render_kwargs: dict[str, Any] = Field(default_factory=dict)

    create: Optional[Callable[..., "ChatCompletion"]] = Field(default=None)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> Union[T, Awaitable[T]]:
        if self.fn is None:
            raise NotImplementedError

        from cyberchipped import settings

        logger = get_logger("cyberchipped.ai_fn")

        logger.debug_kv(
            "AI Function Call",
            f"Calling {self.fn.__name__} with {args} and {kwargs}",
            "blue",
        )

        is_async_fn = asyncio.iscoroutinefunction(self.fn)

        call = "async_call" if is_async_fn else "sync_call"
        create = (
            self.create or settings.openai.chat.completions.acreate
            if is_async_fn
            else settings.openai.chat.completions.create
        )

        result = getattr(self, call)(create, *args, **kwargs)

        logger.debug_kv("AI Function Call", f"Returned {result}", "blue")

        return result

    async def async_call(
        self, acreate: Callable[..., Awaitable[Any]], *args: P.args, **kwargs: P.kwargs
    ) -> T:
        _response = await acreate(**self.as_prompt(*args, **kwargs).serialize())
        return self.parse(_response)

    def sync_call(
        self, create: Callable[..., Any], *args: P.args, **kwargs: P.kwargs
    ) -> T:
        _response = create(**self.as_prompt(*args, **kwargs).serialize())
        return self.parse(_response)

    def parse(self, response: "ChatCompletion") -> T:
        tool_calls = response.choices[0].message.tool_calls
        if tool_calls is None:
            raise NotImplementedError
        if self.fn is None:
            raise NotImplementedError
        arguments = tool_calls[0].function.arguments

        tool = create_tool_from_type(
            _type=self.fn.__annotations__["return"],
            model_name=self.name,
            model_description=self.description,
            field_name=self.field_name,
            field_description=self.field_description,
        ).function
        if not tool or not tool.model:
            raise NotImplementedError
        try:
            return getattr(tool.model.model_validate_json(arguments), self.field_name)
        except ValidationError:
            # When the user provides a dict obj as a type hint, the arguments
            # are returned usually as an object and not a nested dict.
            _arguments: str = json.dumps({self.field_name: json.loads(arguments)})
            return getattr(tool.model.model_validate_json(_arguments), self.field_name)

    @expose_sync_method("map")
    async def amap(self, *map_args: list[Any], **map_kwargs: list[Any]) -> list[T]:
        """
        Map the AI function over a sequence of arguments. Runs concurrently.

        A `map` twin method is provided by the `expose_sync_method` decorator.

        You can use `map` or `amap` synchronously or asynchronously, respectively,
        regardless of whether the user function is synchronous or asynchronous.

        Arguments should be provided as if calling the function normally, but
        each argument must be a list. The function is called once for each item
        in the list, and the results are returned in a list.

        For example, fn.map([1, 2]) is equivalent to [fn(1), fn(2)].

        fn.map([1, 2], x=['a', 'b']) is equivalent to [fn(1, x='a'), fn(2, x='b')].
        """
        tasks: list[Any] = []
        if map_args and map_kwargs:
            max_length = max(
                len(arg) for arg in (map_args + tuple(map_kwargs.values()))
            )
        elif map_args:
            max_length = max(len(arg) for arg in map_args)
        else:
            max_length = max(len(v) for v in map_kwargs.values())

        for i in range(max_length):
            call_args = [arg[i] if i < len(arg) else None for arg in map_args]
            call_kwargs = (
                {k: v[i] if i < len(v) else None for k, v in map_kwargs.items()}
                if map_kwargs
                else {}
            )

            tasks.append(run_async(self, *call_args, **call_kwargs))

        return await asyncio.gather(*tasks)

    def as_prompt(
        self,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> PromptFunction[BaseModel]:
        return PromptFunction[BaseModel].as_function_call(
            fn=self.fn,
            environment=self.environment,
            prompt=self.prompt,
            model_name=self.name,
            model_description=self.description,
            field_name=self.field_name,
            field_description=self.field_description,
            **self.render_kwargs,
        )(*args, **kwargs)

    @overload
    @classmethod
    def as_decorator(
        cls: type[Self],
        *,
        environment: Optional[BaseEnvironment] = None,
        prompt: Optional[str] = None,
        model_name: str = "FormatResponse",
        model_description: str = "Formats the response.",
        field_name: str = "data",
        field_description: str = "The data to format.",
        acreate: Optional[Callable[..., Awaitable[Any]]] = None,
        **render_kwargs: Any,
    ) -> Callable[P, Self]:
        pass

    @overload
    @classmethod
    def as_decorator(
        cls: type[Self],
        fn: Callable[P, T],
        *,
        environment: Optional[BaseEnvironment] = None,
        prompt: Optional[str] = None,
        model_name: str = "FormatResponse",
        model_description: str = "Formats the response.",
        field_name: str = "data",
        field_description: str = "The data to format.",
        acreate: Optional[Callable[..., Awaitable[Any]]] = None,
        **render_kwargs: Any,
    ) -> Self:
        pass

    @classmethod
    def as_decorator(
        cls: type[Self],
        fn: Optional[Callable[P, T]] = None,
        *,
        environment: Optional[BaseEnvironment] = None,
        prompt: Optional[str] = None,
        model_name: str = "FormatResponse",
        model_description: str = "Formats the response.",
        field_name: str = "data",
        field_description: str = "The data to format.",
        **render_kwargs: Any,
    ) -> Union[Callable[[Callable[P, T]], Self], Self]:
        def decorator(func: Callable[P, T]) -> Self:
            return cls(
                fn=func,
                environment=environment,
                name=model_name,
                description=model_description,
                field_name=field_name,
                field_description=field_description,
                **({"prompt": prompt} if prompt else {}),
                **render_kwargs,
            )

        if fn is not None:
            return decorator(fn)

        return decorator


@overload
def ai_fn(
    *,
    environment: Optional[BaseEnvironment] = None,
    prompt: Optional[str] = None,
    model_name: str = "FormatResponse",
    model_description: str = "Formats the response.",
    field_name: str = "data",
    field_description: str = "The data to format.",
    **render_kwargs: Any,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    pass


@overload
def ai_fn(
    fn: Callable[P, T],
    *,
    environment: Optional[BaseEnvironment] = None,
    prompt: Optional[str] = None,
    model_name: str = "FormatResponse",
    model_description: str = "Formats the response.",
    field_name: str = "data",
    field_description: str = "The data to format.",
    **render_kwargs: Any,
) -> Callable[P, T]:
    pass


def ai_fn(
    fn: Optional[Callable[P, T]] = None,
    *,
    environment: Optional[BaseEnvironment] = None,
    prompt: Optional[str] = None,
    model_name: str = "FormatResponse",
    model_description: str = "Formats the response.",
    field_name: str = "data",
    field_description: str = "The data to format.",
    **render_kwargs: Any,
) -> Union[Callable[[Callable[P, T]], Callable[P, T]], Callable[P, T]]:
    if fn is not None:
        return AIFunction.as_decorator(  # type: ignore
            fn=fn,
            environment=environment,
            prompt=prompt,
            model_name=model_name,
            model_description=model_description,
            field_name=field_name,
            field_description=field_description,
            **render_kwargs,
        )

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        return AIFunction.as_decorator(  # type: ignore
            fn=func,
            environment=environment,
            prompt=prompt,
            model_name=model_name,
            model_description=model_description,
            field_name=field_name,
            field_description=field_description,
            **render_kwargs,
        )

    return decorator



import logging
from functools import lru_cache, partial
from typing import Optional

from rich.logging import RichHandler  # type: ignore
from rich.markup import escape  # type: ignore

import cyberchipped


@lru_cache()
def get_logger(
    name: Optional[str] = None,
) -> logging.Logger:
    parent_logger = logging.getLogger("cyberchipped")

    if name:
        # Append the name if given but allow explicit full names e.g. "cyberchipped.test"
        # should not become "cyberchipped.cyberchipped.test"
        if not name.startswith(parent_logger.name + "."):
            logger = parent_logger.getChild(name)
        else:
            logger = logging.getLogger(name)
    else:
        logger = parent_logger

    add_logging_methods(logger)
    return logger


def setup_logging(
    level: Optional[str] = None,
) -> None:
    logger = get_logger()

    if level is not None:
        logger.setLevel(level)
    else:
        logger.setLevel(cyberchipped.settings.log_level)

    logger.handlers.clear()

    handler = RichHandler(
        rich_tracebacks=True,
        markup=False,
        # console=Console(width=cyberchipped.settings.log_console_width),
    )
    formatter = logging.Formatter("%(name)s: %(message)s")
    handler.setFormatter(formatter)

    logger.addHandler(handler)
    logger.propagate = False


def add_logging_methods(logger: logging.Logger) -> None:
    def log_style(level: int, message: str, style: Optional[str] = None):
        if not style:
            style = "default on default"
        message = f"[{style}]{escape(str(message))}[/]"
        logger.log(level, message, extra={"markup": True})

    def log_kv(
        level: int,
        key: str,
        value: str,
        key_style: str = "default on default",
        value_style: str = "default on default",
        delimiter: str = ": ",
    ):
        logger.log(
            level,
            f"[{key_style}]{escape(str(key))}{delimiter}[/][{value_style}]{escape(str(value))}[/]",
            extra={"markup": True},
        )

    setattr(logger, "debug_style", partial(log_style, logging.DEBUG))
    setattr(logger, "info_style", partial(log_style, logging.INFO))
    setattr(logger, "warning_style", partial(log_style, logging.WARNING))
    setattr(logger, "error_style", partial(log_style, logging.ERROR))
    setattr(logger, "critical_style", partial(log_style, logging.CRITICAL))

    setattr(logger, "debug_kv", partial(log_kv, logging.DEBUG))
    setattr(logger, "info_kv", partial(log_kv, logging.INFO))
    setattr(logger, "warning_kv", partial(log_kv, logging.WARNING))
    setattr(logger, "error_kv", partial(log_kv, logging.ERROR))
    setattr(logger, "critical_kv", partial(log_kv, logging.CRITICAL))


setup_logging(level=cyberchipped.settings.log_level)



from .settings import settings

from .assistants import Assistant

from .components import ai_fn, ai_model, ai_classifier

__all__ = [
    "ai_fn",
    "ai_model",
    "ai_classifier",
    "settings",
    "Assistant",
]



import tiktoken


def tokenize(text: str, model: str = "gpt-3.5-turbo-1106") -> list[int]:
    tokenizer = tiktoken.encoding_for_model(model)
    return tokenizer.encode(text)


def detokenize(tokens: list[int], model: str = "gpt-3.5-turbo-1106") -> str:
    tokenizer = tiktoken.encoding_for_model(model)
    return tokenizer.decode(tokens)


def count_tokens(text: str) -> int:
    return len(tokenize(text))


def slice_tokens(text: str, n_tokens: int) -> str:
    tokens = tokenize(text)
    return detokenize(tokens[:n_tokens])



import os
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from pydantic import Field, SecretStr
from pydantic_settings import BaseSettings, SettingsConfigDict

if TYPE_CHECKING:
    from openai import AsyncClient, Client
    from openai.types.chat import ChatCompletion


class Settings(BaseSettings):
    model_config = SettingsConfigDict(
        env_prefix="cyberchipped_",
        env_file="~/.cyberchipped/.env",
        extra="allow",
        arbitrary_types_allowed=True,
    )

    def __setattr__(self, name: str, value: Any) -> None:
        """Preserve SecretStr type when setting values."""
        field = self.model_fields.get(name)
        if field:
            annotation = field.annotation
            base_types = (
                annotation.__args__
                if getattr(annotation, "__origin__", None) is Union
                else (annotation,)
            )
            if SecretStr in base_types and not isinstance(value, SecretStr):
                value = SecretStr(value)
        super().__setattr__(name, value)


class ModelSettings(Settings):
    model: str

    @property
    def encoder(self):
        import tiktoken

        return tiktoken.encoding_for_model(self.model).encode


class ChatCompletionSettings(ModelSettings):
    model: str = Field(
        default="gpt-3.5-turbo-1106",
        description="The default chat model to use.",
    )

    async def acreate(self, **kwargs: Any) -> "ChatCompletion":
        from cyberchipped.settings import settings

        return await settings.openai.async_client.chat.completions.create(
            model=self.model, **kwargs
        )

    def create(self, **kwargs: Any) -> "ChatCompletion":
        from cyberchipped.settings import settings

        return settings.openai.client.chat.completions.create(
            model=self.model, **kwargs
        )


class AssistantSettings(ModelSettings):
    model: str = Field(
        default="gpt-4-1106-preview",
        description="The default assistant model to use.",
    )


class ChatSettings(Settings):
    completions: ChatCompletionSettings = Field(
        default_factory=ChatCompletionSettings)


class OpenAISettings(Settings):
    model_config = SettingsConfigDict(env_prefix="_openai_")

    api_key: Optional[SecretStr] = Field(
        default=None,
        description="Your OpenAI API key.",
    )

    organization: Optional[str] = Field(
        default=None,
        description="Your OpenAI organization ID.",
    )

    chat: ChatSettings = Field(default_factory=ChatSettings)
    assistants: AssistantSettings = Field(default_factory=AssistantSettings)

    @property
    def async_client(
        self, api_key: Optional[str] = None, **kwargs: Any
    ) -> "AsyncClient":
        from openai import AsyncClient

        if not (api_key or self.api_key):
            raise ValueError("No API key provided.")
        elif not api_key and self.api_key:
            api_key = self.api_key.get_secret_value()

        return AsyncClient(
            api_key=api_key,
            organization=self.organization,
            **kwargs,
        )

    @property
    def client(self, api_key: Optional[str] = None, **kwargs: Any) -> "Client":
        from openai import Client

        if not (api_key or self.api_key):
            raise ValueError("No API key provided.")
        elif not api_key and self.api_key:
            api_key = self.api_key.get_secret_value()

        return Client(
            api_key=api_key,
            organization=self.organization,
            **kwargs,
        )


class Settings(Settings):
    model_config = SettingsConfigDict(env_prefix="_")

    openai: OpenAISettings = Field(default_factory=OpenAISettings)

    log_level: str = Field(
        default="DEBUG",
        description="The log level to use.",
    )


settings = Settings()


@contextmanager
def temporary_settings(**kwargs: Any):
    """
    Temporarily override cyberchipped setting values. This will _not_ mutate values that have
    been already been accessed at module load time.

    This function should only be used for testing.
    """
    old_env = os.environ.copy()
    old_settings = settings.model_copy()

    try:
        for setting in kwargs:
            value = kwargs.get(setting)
            if value is not None:
                os.environ[setting] = str(value)
            else:
                os.environ.pop(setting, None)

        new_settings = Settings()

        for field in settings.model_fields:
            object.__setattr__(settings, field, getattr(new_settings, field))

        yield settings
    finally:
        for setting in kwargs:
            value = old_env.get(setting)
            if value is not None:
                os.environ[setting] = value
            else:
                os.environ.pop(setting, None)

        for field in settings.model_fields:
            object.__setattr__(settings, field, getattr(old_settings, field))



import tempfile
from datetime import datetime

import openai
from openai.types.beta.threads import ThreadMessage
from openai.types.beta.threads.runs.run_step import RunStep
from rich import box
from rich.console import Console
from rich.panel import Panel

# def pprint_run(run: Run):
#     """
#     Runs are comprised of steps and messages, which are each in a sorted list
#     BUT the created_at timestamps only have second-level resolution, so we can't
#     easily sort the lists. Instead we walk them in order and combine them giving
#     ties to run steps.
#     """
#     index_steps = 0
#     index_messages = 0
#     combined = []

#     while index_steps < len(run.steps) and index_messages < len(run.messages):
#         if (run.steps[index_steps].created_at
#             <= run.messages[index_messages].created_at):
#             combined.append(run.steps[index_steps])
#             index_steps += 1
#         elif (
#             run.steps[index_steps].created_at
#             > run.messages[index_messages].created_at
#         ):
#             combined.append(run.messages[index_messages])
#             index_messages += 1

#     # Add any remaining items from either list
#     combined.extend(run.steps[index_steps:])
#     combined.extend(run.messages[index_messages:])

#     for obj in combined:
#         if isinstance(obj, RunStep):
#             pprint_run_step(obj)
#         elif isinstance(obj, ThreadMessage):
#             pprint_message(obj)


def pprint_run_step(run_step: RunStep):
    # Timestamp formatting
    timestamp = datetime.fromtimestamp(run_step.created_at).strftime("%l:%M:%S %p")

    # default content
    content = (
        f"Assistant is performing an action: {run_step.type} - Status:"
        f" {run_step.status}"
    )

    # attempt to customize content
    if run_step.type == "tool_calls":
        for tool_call in run_step.step_details.tool_calls:
            if tool_call.type == "code_interpreter":
                if run_step.status == "in_progress":
                    content = "Assistant is running the code interpreter..."
                elif run_step.status == "completed":
                    content = "Assistant ran the code interpreter."
                else:
                    content = f"Assistant code interpreter status: {run_step.status}"
            elif tool_call.type == "function":
                if run_step.status == "in_progress":
                    content = (
                        "Assistant used the tool"
                        f" `{tool_call.function.name}` with arguments"
                        f" {tool_call.function.arguments}..."
                    )
                elif run_step.status == "completed":
                    content = (
                        "Assistant used the tool"
                        f" `{tool_call.function.name}` with arguments"
                        f" {tool_call.function.arguments}."
                    )
                else:
                    content = (
                        f"Assistant tool `{tool_call.function.name}` status:"
                        f" `{run_step.status}`"
                    )
    elif run_step.type == "message_creation":
        return

    console = Console()

    # Create the panel for the run step status
    panel = Panel(
        content.strip(),
        title="Assistant Run Step",
        subtitle=f"[italic]{timestamp}[/]",
        title_align="left",
        subtitle_align="right",
        border_style="gray74",
        box=box.ROUNDED,
        width=100,
        expand=True,
        padding=(0, 1),
    )
    # Printing the panel
    console.print(panel)


def download_temp_file(file_id: str, suffix: str = None):
    client = openai.Client()
    # file_info = client.files.retrieve(file_id)
    file_content_response = client.files.with_raw_response.retrieve_content(file_id)

    # Create a temporary file with a context manager to ensure it's cleaned up
    # properly
    with tempfile.NamedTemporaryFile(
        delete=False, mode="wb", suffix=f"{suffix}"
    ) as temp_file:
        temp_file.write(file_content_response.content)
        temp_file_path = temp_file.name  # Save the path of the temp file

    return temp_file_path


def pprint_message(message: ThreadMessage):
    """
    Pretty-prints a single message using the rich library, highlighting the
    speaker's role, the message text, any available images, and the message
    timestamp in a panel format.

    Args:
    message (dict): A message object as described in the API documentation.
    """
    console = Console()
    role_colors = {
        "user": "green",
        "assistant": "blue",
    }

    color = role_colors.get(message.role, "red")
    timestamp = datetime.fromtimestamp(message.created_at).strftime("%l:%M:%S %p")

    content = ""
    for item in message.content:
        if item.type == "text":
            content += item.text.value + "\n\n"
        elif item.type == "image_file":
            # Use the download_temp_file function to download the file and get
            # the local path
            local_file_path = download_temp_file(item.image_file.file_id, suffix=".png")
            # Add a clickable hyperlink to the content
            file_url = f"file://{local_file_path}"
            content += (
                "[bold]Attachment[/bold]:"
                f" [blue][link={file_url}]{local_file_path}[/link][/blue]\n\n"
            )

    for file_id in message.file_ids:
        content += f"Attached file: {file_id}\n"

    # Create the panel for the message
    panel = Panel(
        content.strip(),
        title=f"[bold]{message.role.capitalize()}[/]",
        subtitle=f"[italic]{timestamp}[/]",
        title_align="left",
        subtitle_align="right",
        border_style=color,
        box=box.ROUNDED,
        # highlight=True,
        width=100,  # Fixed width for all panels
        expand=True,  # Panels always expand to the width of the console
        padding=(1, 2),
    )

    # Printing the panel
    console.print(panel)


def pprint_messages(messages: list[ThreadMessage]):
    for message in messages:
        pprint_message(message)



from enum import Enum
from types import GenericAlias
from typing import (
    Any,
    Callable,
    Literal,
    Optional,
    TypeVar,
    Union,
    get_args,
    get_origin,
)

from pydantic import BaseModel, create_model
from pydantic.fields import FieldInfo
from pydantic.json_schema import GenerateJsonSchema, JsonSchemaMode

from cyberchipped import settings
from cyberchipped.requests import Function, Grammar, Tool

U = TypeVar("U", bound=BaseModel)


class FunctionSchema(GenerateJsonSchema):
    def generate(self, schema: Any, mode: JsonSchemaMode = "validation"):
        json_schema = super().generate(schema, mode=mode)
        json_schema.pop("title", None)
        return json_schema


def create_tool_from_type(
    _type: Union[type, GenericAlias],
    model_name: str,
    model_description: str,
    field_name: str,
    field_description: str,
    python_function: Optional[Callable[..., Any]] = None,
    **kwargs: Any,
) -> Tool[BaseModel]:
    annotated_metadata = getattr(_type, "__metadata__", [])
    if isinstance(next(iter(annotated_metadata), None), FieldInfo):
        metadata = next(iter(annotated_metadata))
    else:
        metadata = FieldInfo(description=field_description)

    model: type[BaseModel] = create_model(
        model_name,
        __config__=None,
        __base__=None,
        __module__=__name__,
        __validators__=None,
        __cls_kwargs__=None,
        **{field_name: (_type, metadata)},
    )
    return Tool[BaseModel](
        type="function",
        function=Function[BaseModel](
            name=model_name,
            description=model_description,
            parameters=model.model_json_schema(schema_generator=FunctionSchema),
            model=model,
        ),
    )


def create_vocabulary_from_type(
    vocabulary: Union[GenericAlias, type, list[str]],
) -> list[str]:
    if get_origin(vocabulary) == Literal:
        return [str(token) for token in get_args(vocabulary)]
    elif isinstance(vocabulary, type) and issubclass(vocabulary, Enum):
        return [str(token) for token in list(vocabulary.__members__.keys())]
    elif isinstance(vocabulary, list) and next(iter(get_args(list[str])), None) == str:
        return [str(token) for token in vocabulary]
    else:
        raise TypeError(
            f"Expected Literal or Enum or list[str], got {type(vocabulary)} with value"
            f" {vocabulary}"
        )


def create_grammar_from_vocabulary(
    vocabulary: list[str],
    encoder: Callable[[str], list[int]] = settings.openai.chat.completions.encoder,
    max_tokens: Optional[int] = None,
    _enumerate: bool = True,
    **kwargs: Any,
) -> Grammar:
    return Grammar(
        max_tokens=max_tokens,
        logit_bias={
            str(encoding): 100
            for i, token in enumerate(vocabulary)
            for encoding in encoder(str(i) if _enumerate else token)
        },
    )



import asyncio
from typing import Any, Callable, Optional, Union

from openai.types.beta.threads.run import Run as OpenAIRun
from openai.types.beta.threads.runs import RunStep as OpenAIRunStep
from pydantic import BaseModel, Field, PrivateAttr, field_validator

import cyberchipped.utilities.tools
from cyberchipped.requests import Tool
from cyberchipped.tools.assistants import AssistantTools, CancelRun
from cyberchipped.utilities.asyncio import ExposeSyncMethodsMixin, expose_sync_method
from cyberchipped.utilities.logging import get_logger
from cyberchipped.utilities.openai import get_client

from .assistants import Assistant
from .threads import Thread

logger = get_logger("Runs")


class Run(BaseModel, ExposeSyncMethodsMixin):
    thread: Thread
    assistant: Assistant
    instructions: Optional[str] = Field(
        None, description="Replacement instructions to use for the run."
    )
    additional_instructions: Optional[str] = Field(
        None,
        description=(
            "Additional instructions to append to the assistant's instructions."
        ),
    )
    tools: Optional[list[Union[AssistantTools, Callable]]] = Field(
        None, description="Replacement tools to use for the run."
    )
    additional_tools: Optional[list[AssistantTools]] = Field(
        None,
        description="Additional tools to append to the assistant's tools. ",
    )
    run: OpenAIRun = None
    data: Any = None

    @field_validator("tools", "additional_tools", mode="before")
    def format_tools(cls, tools: Union[None, list[Union[Tool, Callable]]]):
        if tools is not None:
            return [
                (
                    tool
                    if isinstance(tool, Tool)
                    else cyberchipped.utilities.tools.tool_from_function(tool)
                )
                for tool in tools
            ]

    @expose_sync_method("refresh")
    async def refresh_async(self):
        client = get_client()
        self.run = await client.beta.threads.runs.retrieve(
            run_id=self.run.id, thread_id=self.thread.id
        )

    @expose_sync_method("cancel")
    async def cancel_async(self):
        client = get_client()
        await client.beta.threads.runs.cancel(
            run_id=self.run.id, thread_id=self.thread.id
        )

    async def _handle_step_requires_action(self):
        client = get_client()
        if self.run.status != "requires_action":
            return
        if self.run.required_action.type == "submit_tool_outputs":
            tool_outputs = []
            tools = self.get_tools()

            for tool_call in self.run.required_action.submit_tool_outputs.tool_calls:
                try:
                    output = cyberchipped.utilities.tools.call_function_tool(
                        tools=tools,
                        function_name=tool_call.function.name,
                        function_arguments_json=tool_call.function.arguments,
                    )
                except CancelRun as exc:
                    logger.debug(f"Ending run with data: {exc.data}")
                    raise
                except Exception as exc:
                    output = f"Error calling function {tool_call.function.name}: {exc}"
                    logger.error(output)
                tool_outputs.append(
                    dict(tool_call_id=tool_call.id, output=output or "")
                )

            await client.beta.threads.runs.submit_tool_outputs(
                thread_id=self.thread.id, run_id=self.run.id, tool_outputs=tool_outputs
            )

    def get_instructions(self) -> str:
        if self.instructions is None:
            instructions = self.assistant.get_instructions() or ""
        else:
            instructions = self.instructions

        if self.additional_instructions is not None:
            instructions = "\n\n".join(
                [instructions, self.additional_instructions])

        return instructions

    def get_tools(self) -> list[AssistantTools]:
        tools = []
        if self.tools is None:
            tools.extend(self.assistant.get_tools())
        else:
            tools.extend(self.tools)
        if self.additional_tools is not None:
            tools.extend(self.additional_tools)
        return tools

    async def run_async(self) -> "Run":
        client = get_client()

        create_kwargs = {}

        if self.instructions is not None or self.additional_instructions is not None:
            create_kwargs["instructions"] = self.get_instructions()

        if self.tools is not None or self.additional_tools is not None:
            create_kwargs["tools"] = self.get_tools()

        self.run = await client.beta.threads.runs.create(
            thread_id=self.thread.id, assistant_id=self.assistant.id, **create_kwargs
        )

        try:
            await asyncio.wait_for(self._run_loop(), timeout=60)
        except asyncio.TimeoutError:
            if self.run.status != "completed":
                # Cancel the run if it's not completed
                await client.beta.threads.runs.cancel(
                    run_id=self.run.id, thread_id=self.thread.id
                )
                self.data = "Run cancelled due to timeout."
            else:
                self.data = "Run already completed; no need to cancel."
            await self.refresh_async()

    async def _run_loop(self):
        while self.run.status in ("queued", "in_progress", "requires_action"):
            if self.run.status == "requires_action":
                await self._handle_step_requires_action()
            await asyncio.sleep(0.1)
            await self.refresh_async()

class RunMonitor(BaseModel):
    run_id: str
    thread_id: str
    _run: Run = PrivateAttr()
    _thread: Thread = PrivateAttr()
    steps: list[OpenAIRunStep] = []

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._thread = Thread(**kwargs["thread_id"])
        self._run = Run(**kwargs["run_id"], thread=self.thread)

    @property
    def thread(self):
        return self._thread

    @property
    def run(self):
        return self._run

    async def refresh_run_steps_async(self):
        """
        Asynchronously refreshes and updates the run steps list.

        This function fetches the latest run steps up to a specified limit and
        checks if the latest run step in the current run steps list
        (`self.steps`) is included in the new batch. If the latest run step is
        missing, it continues to fetch additional run steps in batches, up to a
        maximum count, using pagination. The function then updates
        `self.steps` with these new run steps, ensuring any existing run steps
        are updated with their latest versions and new run steps are appended in
        their original order.
        """
        # fetch up to 100 run steps
        max_fetched = 100
        limit = 50
        max_attempts = max_fetched / limit + 2

        # Fetch the latest run steps
        client = get_client()

        response = await client.beta.threads.runs.steps.list(
            run_id=self.run.id,
            thread_id=self.thread.id,
            limit=limit,
        )
        run_steps = list(reversed(response.data))

        if not run_steps:
            return

        # Check if the latest run step in self.steps is in the new run steps
        latest_step_id = self.steps[-1].id if self.steps else None
        missing_latest = (
            latest_step_id not in {rs.id for rs in run_steps}
            if latest_step_id
            else True
        )

        # If the latest run step is missing, fetch additional run steps
        total_fetched = len(run_steps)
        attempts = 0
        while (
            run_steps
            and missing_latest
            and total_fetched < max_fetched
            and attempts < max_attempts
        ):
            attempts += 1
            response = await client.beta.threads.runs.steps.list(
                run_id=self.run.id,
                thread_id=self.thread.id,
                limit=limit,
                # because this is a raw API call, "after" refers to pagination
                # in descnding chronological order
                after=run_steps[0].id,
            )
            paginated_steps = list(reversed(response.data))

            total_fetched += len(paginated_steps)
            # prepend run steps
            run_steps = paginated_steps + run_steps
            if any(rs.id == latest_step_id for rs in paginated_steps):
                missing_latest = False

        # Update self.steps with the latest data
        new_steps_dict = {rs.id: rs for rs in run_steps}
        for i in range(len(self.steps) - 1, -1, -1):
            if self.steps[i].id in new_steps_dict:
                self.steps[i] = new_steps_dict.pop(self.steps[i].id)
            else:
                break
        # Append remaining new run steps at the end in their original order
        self.steps.extend(new_steps_dict.values())



import inspect
import re
from datetime import datetime
from typing import Any, ClassVar, Pattern, Union
from zoneinfo import ZoneInfo

import pydantic
from jinja2 import Environment as JinjaEnvironment
from jinja2 import StrictUndefined, select_autoescape
from jinja2 import Template as BaseTemplate
from typing_extensions import Self

from cyberchipped.requests import BaseMessage as Message


class BaseEnvironment(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    environment: JinjaEnvironment = pydantic.Field(
        default=JinjaEnvironment(
            autoescape=select_autoescape(default_for_string=False),
            trim_blocks=True,
            lstrip_blocks=True,
            auto_reload=True,
            undefined=StrictUndefined,
        )
    )

    globals: dict[str, Any] = pydantic.Field(
        default_factory=lambda: {
            "now": lambda: datetime.now(ZoneInfo("UTC")),
            "inspect": inspect,
        }
    )

    @pydantic.model_validator(mode="after")
    def setup_globals(self: Self) -> Self:
        self.environment.globals.update(self.globals)  # type: ignore
        return self

    def render(self, template: Union[str, BaseTemplate], **kwargs: Any) -> str:
        if isinstance(template, str):
            return self.environment.from_string(template).render(**kwargs)
        return template.render(**kwargs)


Environment = BaseEnvironment()


def split_text_by_tokens(
    text: str,
    split_tokens: list[str],
    environment: JinjaEnvironment = Environment.environment,
) -> list[tuple[str, str]]:
    cleaned_text = inspect.cleandoc(text)

    # Find all positions of tokens in the text
    positions = [
        (match.start(), match.end(), match.group().rstrip(":").strip())
        for token in split_tokens
        for match in re.finditer(re.escape(token) + r"(?::\s*)?", cleaned_text)
    ]

    # Sort positions by their start index
    positions.sort(key=lambda x: x[0])

    paired: list[tuple[str, str]] = []
    prev_end = 0
    prev_token = split_tokens[0]
    for start, end, token in positions:
        paired.append((prev_token, cleaned_text[prev_end:start].strip()))
        prev_end = end
        prev_token = token

    paired.append((prev_token, cleaned_text[prev_end:].strip()))

    # Remove pairs where the text is empty
    paired = [(token.replace(":", ""), text) for token, text in paired if text]

    return paired


class Transcript(pydantic.BaseModel):
    content: str
    roles: list[str] = pydantic.Field(default=["system", "user"])
    environment: ClassVar[BaseEnvironment] = Environment

    @property
    def role_regex(self) -> Pattern[str]:
        return re.compile("|".join([f"\n\n{role}:" for role in self.roles]))

    def render(self: Self, **kwargs: Any) -> str:
        return self.environment.render(self.content, **kwargs)

    def render_to_messages(
        self: Self,
        **kwargs: Any,
    ) -> list[Message]:
        pairs = split_text_by_tokens(
            text=self.render(**kwargs),
            split_tokens=[f"\n{role}" for role in self.roles],
        )
        return [
            Message(
                role=pair[0].strip(),
                content=pair[1],
            )
            for pair in pairs
        ]
